package it.polito.bigdata.spark.exercise2;
import org.apache.spark.api.java.*;
import scala.Tuple2;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.apache.spark.SparkConf;

public class SparkDriver {

    public static void main(String[] args) {
        
		Logger.getLogger("org").setLevel(Level.OFF);
		Logger.getLogger("akka").setLevel(Level.OFF);

        String purchasePath = "data/Purchases.txt";
        //String usersPath = "data/Users.txt";
        String cataloguePath = "data/Catalogue.txt";
        String outputPath1;
        String outputPath2;

        
        outputPath1 = "outPart1/";
        outputPath2 = "outPart2/";

        // Initialize SparkContext
        SparkConf conf = new SparkConf().setAppName("Exam 20240219_Spark").setMaster("local");
		// Create a Spark Context object
		JavaSparkContext sc = new JavaSparkContext(conf);


        // load all rdds
        // SaleTimestamp,UserID,ItemID,SalePrice
        JavaRDD<String> purchaseRDD = sc.textFile(purchasePath);
        // UserID,Name,Surname,City,Country
        // JavaRDD<String> usersRDD = sc.textFile(usersPath);
        // ItemID,Name,Category,StillInProduction
        JavaRDD<String> catalogueRDD = sc.textFile(cataloguePath);

        // ########################################
        // PART 1
        // ########################################
        // filter and keep only the purchases in year 2022 and year 2023
        JavaRDD<String> purchases2223 = purchaseRDD
                                        .filter(s ->
                                            s.startsWith("2023") || 
                                            s.startsWith("2022"))
                                        .cache();

        // count for each user and year the number of purchases the user made
        // key = userId
        // value = (count2022, count2023)
        // and use a reduceByKey to count the number of purchases for each year 2002 and 2023 separately
        JavaPairRDD<String, Tuple2<Integer, Integer>> userCountPurchases2223 = 
                    purchases2223.mapToPair(s -> {
                        String fields[] = s.split(",");
                        String userId = fields[1];
                        String year = fields[0].split("/")[0];
                        if( year.equals("2022") )
                            return new Tuple2<String, Tuple2<Integer, Integer>>(
                                userId, new Tuple2<>(1,0)   // add 1 to count2022, 0 to count2023
                                );
                        else
                            return new Tuple2<String, Tuple2<Integer, Integer>>(
                                userId, new Tuple2<>(0,1)   // add 1 to count2023, 0 to count2022
                                );
                    })
                    .reduceByKey((a,b)-> new Tuple2<Integer, Integer>(
                        a._1()+b._1(),      // total count 2022 
                        a._2()+b._2()))     // total count 2023
                    .cache();

        // compute for each year 2022 and 2023, the maximum number of purchases among all users
        Tuple2<Integer, Integer> maxPurchases2223 = 
                userCountPurchases2223.values().reduce(
                        (a,b) -> 
                        new Tuple2<Integer, Integer>(
                            (a._1() > b._1()) ? a._1() : b._1(), // max 2022 
                            (a._2() > b._2()) ? a._2() : b._2()  // max 2023
                            )
                );
        int max22 = maxPurchases2223._1();
        int max23 = maxPurchases2223._2();

        // filter the JavaPairRDD userCountPurchases2223 
        // by keeping only the records with value part (count of purchases) equal to max22 or max23
        // and then return only the unique users (keys)
        JavaRDD<String> res1 = userCountPurchases2223
                .filter(x -> (x._2()._1() == max22 || x._2()._2() == max23) ? true : false)
                .keys();

        res1.saveAsTextFile(outputPath1);
   
        

        // ########################################
        // PART 2 v1
        // ########################################
        // considering the purchases in year 2022/2023 (purchases2223 RDD)
        // we use a mapToPair with
        // key = itemID
        // value = userID
        // and a distinct to obtain the distinct user-product purchases,
        // a perform a mapToPair + reduceByKey to count for each itemID,
        // the number of distinct users who bought that item
        // key = itemID
        // value = numberOfDistinctUsersPurchases
        JavaPairRDD<String, Integer> itemDistinctUsersPurchases = purchases2223
                            .mapToPair(s -> {
                                String fields[] = s.split(",");
                                String userId = fields[1];
                                String itemId = fields[2];
                                return new Tuple2<>(itemId, userId);
                            })
                            .distinct()
                            .mapToPair(t -> new Tuple2<>(t._1(), 1))
                            .reduceByKey((v1, v2) -> v1 + v2);

        // for each item, we retrieve the corresponding category
        JavaPairRDD<String, String> itemCategory = catalogueRDD
                            .mapToPair(s -> {
                                String[] fields = s.split(",");
                                String itemId = fields[0];
                                String category = fields[2];
                                return new Tuple2<>(itemId, category);
                            })
                            .cache();

        // we join itemCategory RDD with itemDistinctUsersPurchases
        JavaPairRDD<String, Tuple2<String, Integer>> itemCategoryPurchases = itemCategory
                            .join(itemDistinctUsersPurchases)
                            .cache();

        // compute for each category the maximum number distinct users who purchased the item
        // first, we obtain the following pairRDD
        // key = category
        // value = number of distinct users who purchased an item
        // and then we use a reduceByKey to compute the maximum value for each category
        JavaPairRDD<String, Integer> maxDistinctUsersPurchasesPerCategory = itemCategoryPurchases
                            .mapToPair(t -> new Tuple2<>(t._2()._1(), t._2()._2()))
                            .reduceByKey((v1, v2) -> v1 > v2 ? v1 : v2);

        // join itemCategoryPurchases with maxDistinctUsersPurchasesPerCategory
        // and filter, keeping only the entries with count == maxCount
        // after join, format is
        // key = category
        // value = (itemId, count), maxCountPerCategory
        // then, use a mapToPair to obtain the format for the result
        // key = category
        // value = itemId
        JavaPairRDD<String, String> res2 = itemCategoryPurchases
                            .mapToPair(t -> new Tuple2<>(t._2()._1(), new Tuple2<>(t._1(), t._2()._2())))
                            .join(maxDistinctUsersPurchasesPerCategory)
                            .filter(t -> t._2()._1()._2() == t._2()._2())
                            .mapToPair(t -> {
                                String category = t._1();
                                String value = t._2()._1()._1();
                                return new Tuple2<>(category, value);
                            });

        // from res2 we need to add the 0-case, i.e., categories with items which were never purchased
        // we do this by considering all items and categories in catalogue
        // and use a subtractByKey with res2
        // obtaining the following pairRDD
        // key = category
        // value = "NoPurchases"
        JavaPairRDD<String, String> unsoldCategories = itemCategory
                            .mapToPair(t -> t.swap())
                            .subtractByKey(res2)
                            .mapToPair(t -> new Tuple2<>(t._1(), "NoPurchases"));

        // update res2 with a final Union
        JavaPairRDD<String, String> res2Final = res2.union(unsoldCategories);

        res2Final.saveAsTextFile(outputPath2);

        // Stop SparkContext
        sc.close();
    }
}
